/*
 * Copyright (c) 2024, Perforce Software, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <limits.h>

#if defined( OS_NT )
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef GetMessage
#else
#include <sys/time.h>
#endif

#define NEED_GETPID

#include "clientapi.h"
#include "errorlog.h"
#include "errornum.h"
#include "hostenv.h"
#include "ident.h"
#include "options.h"
#include "pathsys.h"
#include "spec.h"
#include "strtable.h"
#include "vararray.h"

#if defined( OS_NT )
#define getprocid() GetCurrentThreadId()
#else
#define getprocid() getpid()
#endif

#if defined( OS_AIX ) || defined( OS_HPUX ) || defined( OS_SOLARIS )
#define srandomL srandom
#define randomL random
#define RANDOM_MAXL INT_MAX
#elif defined( OS_NT )
#define srandomL srand
#define randomL rand
#define RANDOM_MAXL RAND_MAX
#else
#define srandomL srandom
#define randomL random
#define RANDOM_MAXL RAND_MAX
#endif

#define ES_BROWSECHILD 64

Ident ident = {
	IdentMagic "BROWSECHILD" "/" ID_OS "/" ID_REL "/" ID_PATCH,
	ID_Y "/" ID_M "/" ID_D
};

class MsgBrowseChild {

    public:

	static ErrorId ShortUsage;
	static ErrorId BrowsesNotNumeric;
	static ErrorId InvalidConnectType;
	static ErrorId SeedNotNumeric;

} ;

/*
 * Define messages for this application.
 */

ErrorId MsgBrowseChild::ShortUsage		= { ErrorOf( ES_BROWSECHILD, 1, E_FAILED, EV_USAGE, 0 ), "Usage: browsechild [-p port] [-u username] \\\n\t                   [-n browses] [-c {1,b,c}] [-s seed] [-d]\n\t       browsechild -V" } ;

ErrorId MsgBrowseChild::BrowsesNotNumeric	= { ErrorOf( ES_BROWSECHILD, 2, E_FAILED, EV_USAGE, 0 ), "Number of browses must be numeric!" } ;

ErrorId MsgBrowseChild::InvalidConnectType	= { ErrorOf( ES_BROWSECHILD, 3, E_FAILED, EV_USAGE, 0 ), "Connection type must be 1 (single), b (per browse), or c (per command)!" } ;

ErrorId MsgBrowseChild::SeedNotNumeric		= { ErrorOf( ES_BROWSECHILD, 4, E_FAILED, EV_USAGE, 0 ), "Random number seed must be numeric!" } ;

class StrBufArray {

    public:

		~StrBufArray();

	void	Clear();
	int	Count();
	StrPtr*	GetRandom();
	void	Put( StrBuf *strbufp );

    private:

	VarArray strbufs;

} ;

enum ConnectType {
	CONNECT_ONE,
	CONNECT_BROWSE,
	CONNECT_COMMAND
} ;

class OptionsBrowseChild : public Options {

    public:

	StrPtr*	GetPort();
	StrPtr*	GetUsername();
	StrPtr*	GetClientname();
	int	GetBrowses();
	int	GetSeed();

	bool	IsConnectOne();
	bool	IsConnectBrowse();
	bool	IsConnectCommand();
	bool	FstatDc();

	void	SetUserClient();
	void	SetDefaultSeed();
	void	SetFstatDc( StrPtr *slevel );

	void	SetOpts( Error *e );

    private:

	StrPtr* port;		// Perforce server to browse
	StrBuf username;	// name of user as which to browse
	StrBuf clientname;	// name of client to be used while browsing
	int browses;
	ConnectType connect;
	int seed;
	bool fstatDc;		// get directories using fstat -Dc

} ;

class ClientApiBrowseChild : public ClientApi {

    public:

	void	ClearConnectContext();

	void	NewBrowse();
	bool	IsNewBrowse();

	void	Connect( Error *e );
	void	Disconnect( Error *e );
	void	Reconnect( Error *e );

	void	Setup( StrPtr *port, Error *e );
	void	Cleanup( Error *e );

    private:

	bool newbrowse;

} ;

class ClientUserCatchError : public ClientUser {

    public:

	void		Run( OptionsBrowseChild *opts,
			    ClientApiBrowseChild *client, const char *func,
			    Error *e );
	void		Run( OptionsBrowseChild *opts,
			    ClientApiBrowseChild *client, const char *func,
			    int fargc, char **fargv,
			    Error *e );

	virtual void	HandleError( Error *err );

    private:

	Error ecatch;

} ;

class ClientUserQuiet : public ClientUserCatchError {

    public:

	virtual void	OutputInfo( char level, const char *data );

} ;

class ClientUserClient : public ClientUserQuiet {

    public:

	virtual void	OutputStat( StrDict *varList );
	virtual void	InputData( StrBuf *strbuf, Error *e );

    private:

	StrBufDict vars;

} ;

class ClientUserDepots : public ClientUserQuiet {

    public:

	void		SetDepots( StrBufArray *pStrBufArray );

	virtual void	OutputStat( StrDict *varList );

    private:

	StrBufArray *depots;

} ;

class ClientUserDirs : public ClientUserQuiet {

    public:

	void		SetDirs( StrBufArray *pStrBufArray );

	virtual void	OutputStat( StrDict *varList );

    private:

	StrBufArray *dirs;

} ;

class ClientUserFstat : public ClientUserQuiet {

    public:

	void		SetDepotfiles( StrBufArray *pStrBufArray );
	void		SetDirs( StrBufArray *pStrBufArray );

	virtual void	HandleError( Error *err );

	virtual void	OutputStat( StrDict *varList );

    private:

	StrBufArray *depotfiles;
	StrBufArray *dirs;

} ;

class ClientUserFilelog : public ClientUserQuiet {

    public:

	virtual void	HandleError( Error *err );

} ;

int nrandom( int lower, int upper )
{
	/*
	 * Return a pseudo-random number within a lower bound (inclusive) and
	 * an upper bound (exclusive).
	 */
	return (long long)randomL() * ( upper - lower ) /
	    ( (long long)RANDOM_MAXL + 1 ) + lower;

}

StrBufArray::~StrBufArray()
{
	Clear();
}

void StrBufArray::Clear()
{
	int i;

	/*
	 * Delete all the the StrBufs in the array.
	 */
	for( i = 0; i < strbufs.Count(); i++ )
	{
	    delete (StrBuf *)strbufs.Get( i );
	}

	strbufs.Clear();
}

int StrBufArray::Count()
{
	/*
	 * Return the number of StrBufs in the array.
	 */
	return strbufs.Count();
}

StrPtr* StrBufArray::GetRandom()
{
	/*
	 * Return a random StrBuf in the array.
	 */
	return (StrPtr *)strbufs.Get( nrandom( 0, strbufs.Count() ) );
}

void StrBufArray::Put( StrBuf *strbufp )
{
	/*
	 * Append a new StrBuf to the array.
	 */
	strbufs.Put( strbufp );
}

StrPtr *OptionsBrowseChild::GetPort()
{
	return port;
}

StrPtr *OptionsBrowseChild::GetUsername()
{
	return &username;
}

StrPtr *OptionsBrowseChild::GetClientname()
{
	return &clientname;
}

int OptionsBrowseChild::GetBrowses()
{
	return browses;
}

int OptionsBrowseChild::GetSeed()
{
	return seed;
}

bool OptionsBrowseChild::IsConnectOne()
{
	return connect == CONNECT_ONE;
}

bool OptionsBrowseChild::IsConnectBrowse()
{
	return connect == CONNECT_BROWSE;
}

bool OptionsBrowseChild::IsConnectCommand()
{
	return connect == CONNECT_COMMAND;
}

bool OptionsBrowseChild::FstatDc()
{
	return fstatDc;
}

void OptionsBrowseChild::SetUserClient()
{
	HostEnv hostenv;
	StrBuf hostname;

	StrPtr *s;

	StrNum procid;

	/*
	 * Get the name of this (client) host,
	 * ending just before the first "." (if present).
	 */
	hostenv.GetHost( hostname );
	hostname.SetLength( strcspn( hostname.Text(), "." ) );

	if( s = GetValue( 'u', 0 ) )
	{
	    /* 
	     * The name of the user as which to browse
	     * was specified on the command line.
	     */
	    username.Set( s );

	    /*
	     * Assemble the name of the client to be used while browsing
	     * using the name of the user as which to browse and the
	     * name of this host.
	     */
	    clientname << username << "-" << hostname;

	    return;
	}

	/*
	 * The name of the user as which to browse was not specified.
	 */

	/*
	 * Get the (zero-filled) five least-significant digits (if PIDs are
	 * larger than 16 bits) of the identifier for this (client) process.
	 */
	procid.Set( (int)getprocid() % 100000, 5 );

	/*
	 * Assemble the name of the user as which to browse.
	 * The name of this host is appended for uniqueness.
	 */
	username << "user" << procid << "-" << hostname;

	/*
	 * Assemble the name of the client to be used while browsing.
	 * The name of this host is appended for uniqueness.
	 */
	clientname << "client" << procid << "-" << hostname;
}

void OptionsBrowseChild::SetDefaultSeed()
{
#if defined( OS_NT )
	FILETIME ft;
	LARGE_INTEGER li;
	__int64 t;

	/*
	 * To provide a microsecond time increment for the default seed,
	 * the Windows FILETIME structure with GetSystemTimeAsFileTime()
	 * is used. We perform 64-bit math on the high and low parts of
	 * the structure.
	 */
	GetSystemTimeAsFileTime(&ft);
	li.LowPart = ft.dwLowDateTime;
	li.HighPart = ft.dwHighDateTime;
	t = li.QuadPart;	// in 100-nanosecond intervals
	t /= 10;		// in microseconds
	seed = (long)(t % 1000000);
#elif defined( OS_CYGWIN )
	/*
	 * Cygwin's tv_usec resolution returned from gettimeofday is poor,
	 * so use the identifier of this process as the seed. Using the
	 * process identifier is only slightly better since process
	 * identifiers can be quickly reused on Windows. But at least the
	 * process identifiers (and therefore the seeds) will be unique
	 * when a number of children are browsing concurrently.
	 */
	seed = (int)getprocid();
#else
	timeval tv;

	gettimeofday( &tv, 0 );
	seed = tv.tv_usec;	// microseconds part of timeval structure
#endif
}

void OptionsBrowseChild::SetFstatDc( StrPtr *slevel )
{
	if( fstatDc = slevel->Atoi() >= 32 )
	{
	    /*
	     * The server level indicates that fstat -Dc can be used to
	     * retrieve directories, negating the need for a separate dirs
	     * command. Use fstat -Dc to retrieve directories unless
	     * requested to use a separate dirs command.
	     */
	    fstatDc = !(*this)[ 'd' ];
	}
}

void OptionsBrowseChild::SetOpts( Error *e )
{
	StrPtr *s;
	char *text;
	char *endptr;

	/*
	 * Get the Perforce server to be browsed if specified
	 * on the command line.
	 */
	port = GetValue( 'p', 0 );

	/*
	 * Set the name of the user as which to browse and
	 * the name of the client to be used while browsing.
	 */
	SetUserClient();

	if( s = GetValue( 'n', 0 ) )
	{
	    /*
	     * Number of browses was specified on the command line.
	     */
	    browses = strtol( s->Text(), &endptr, 10 );
	    if( s->End() != endptr )
	    {
		/*
		 * Number of browses specified was not numeric.
		 */
		e->Set( MsgBrowseChild::BrowsesNotNumeric );
	    }
	}
	else
	{
	    /*
	     * Number of browses was not specified; use a default of 1024.
	     */
	    browses = 1024;
	}

	if( s = GetValue( 'c', 0 ) )
	{
	    /*
	     * Connection type was specified on the command line.
	     */
	    text = s->Text();
	    if( !strcmp( text, "1" ) )
	    {
		/*
		 * A single connection is maintained for all browses.
		 */
		connect = CONNECT_ONE;
	    }
	    else if( !strcmp( text, "b" ) )
	    {
		/*
		 * Establish a new connection for each browse.
		 */
		connect = CONNECT_BROWSE;
	    }
	    else if( !strcmp( text, "c" ) )
	    {
		/*
		 * Each command is issued on a new connection.
		 */
		connect = CONNECT_COMMAND;
	    }
	    else
	    {
		/*
		 * Connection type specified was not valid.
		 */
		e->Set( MsgBrowseChild::InvalidConnectType );
	    }
	}
	else
	{
	    /*
	     * Connection type was not specified; use a default of
	     * a new connection for each browse.
	     */
	    connect = CONNECT_BROWSE;
	}

	if( s = GetValue( 's', 0 ) )
	{
	    /*
	     * Pseudo-random number generator seed was specified
	     * on the command line.
	     */
	    seed = strtol( s->Text(), &endptr, 10 );
	    if( s->End() != endptr )
	    {
		/*
		 * Seed specified was not numeric.
		 */
		e->Set( MsgBrowseChild::SeedNotNumeric );
	    }
	}
	else
	{
	    /*
	     * Pseudo-random number generator seed was not specified;
	     * use an appropriate default.
	     */
	    SetDefaultSeed();
	}
}

void ClientApiBrowseChild::ClearConnectContext()
{
	newbrowse = false;
}

void ClientApiBrowseChild::NewBrowse()
{
	newbrowse = true;
}

bool ClientApiBrowseChild::IsNewBrowse()
{
	return newbrowse;
}

void ClientApiBrowseChild::Connect( Error *e )
{
	/*
	 * Connect to the server.
	 */
	Init( e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Identify ourselves.
	 */
	SetProg( "browsechild" );
}

void ClientApiBrowseChild::Disconnect( Error *e )
{
	/*
	 * Close the connection to the server.
	 */
	Final( e );
	if( e->Test() )
	{
	    return;
	}
}

void ClientApiBrowseChild::Reconnect( Error *e )
{
	/*
	 * Disconnect from the server.
	 */
	Disconnect( e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Reconnect to the server.
	 */
	Connect( e );
	if( e->Test() )
	{
	    return;
	}
}

void ClientApiBrowseChild::Setup( StrPtr* port, Error *e )
{
	if( port )
	{
	    /*
	     * Perforce server was specified; set the port within the API.
	     */
	    SetPort( port );
	}

	/*
	 * Use tag and specstring protocols.
	 */
	SetProtocol( "tag", "" );
	SetProtocol( "specstring", "" );

	Connect( e );
	if( e->Test() )
	{
	    return;
	}

	ClearConnectContext();	// set newbrowse to false
}

void ClientApiBrowseChild::Cleanup( Error *e )
{
	Disconnect( e );
	if( e->Test() )
	{
	    return;
	}
}

void ClientUserCatchError::Run( OptionsBrowseChild *opts,
	ClientApiBrowseChild *client, const char *func,
	Error *e )
{
	Run( opts, client, func, 0, 0, e );
}

void ClientUserCatchError::Run( OptionsBrowseChild *opts,
	ClientApiBrowseChild *client, const char *func,
	int fargc, char **fargv,
	Error *e )
{
	if( opts->IsConnectBrowse() && client->IsNewBrowse() ||
	    opts->IsConnectCommand() )
	{
	    /*
	     * A new connection is established for each browse and the command
	     * is the start of a new browse, or each command is issued on a
	     * new connection.
	     *
	     * Reconnect to the server before running the command.
	     */
	    client->Reconnect( e );
	    if( e->Test() )
	    {
		return;
	    }
	}

	/*
	 * Run the command with the arguments provided.
	 */
	ecatch.Clear();	// clear any previous ignored errors
	client->SetArgv( fargc, fargv );
	client->Run( func, this );
	*e = ecatch;	// copy any caught errors
	if( e->Test() )
	{
	    return;
	}

	client->ClearConnectContext();	// set newbrowse to false
}

void ClientUserCatchError::HandleError( Error *err )
{
	/*
	 * Catch the error so that the caller of "Run" can see it.
	 */
	ecatch = *err;
}

void ClientUserQuiet::OutputInfo( char level, const char *data )
{
}

void ClientUserClient::OutputStat( StrDict *varList )
{
	vars.CopyVars( *varList );
}

void ClientUserClient::InputData( StrBuf *strbuf, Error *e )
{
	PathSys *pathsys;

	Spec spec;
	SpecDataTable specdatatable( &vars );

	/*
	 * Get the global temporary directory by naming a (text) file in the
	 * global temporary directory and the stripping off the file name
	 * (leaving only the path to the global temporary directory).
	 * Naming the file does not actually create the file.
	 */
	pathsys = PathSys::Create();
	pathsys->Set( FileSys::CreateGlobalTemp( FST_TEXT )->Path() );
	pathsys->ToParent();

	/*
	 * Append the name of the client as a subdirectory of the global
	 * temporary directory. The resulting path is then used as the
	 * client's root.
	 */
	pathsys->SetLocal( *pathsys, *vars.GetVar( "Client" ) );
	vars.ReplaceVar( "Root", pathsys->Text() );
	delete pathsys;

	/*
	 * Decode the client specification from the specification definition.
	 */
	spec.Decode( vars.GetVar( "specdef" ), e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Format the client specification into the returned StrBuf as per
	 * the specification definition and the updated vars StrBufDict.
	 */
	spec.Format( &specdatatable, strbuf );
}
	
void ClientUserDepots::SetDepots( StrBufArray *pStrBufArray )
{
	depots = pStrBufArray;
}

void ClientUserDepots::OutputStat( StrDict *varList )
{
	StrBuf *strbufp;

	if( strcmp( varList->GetVar( "type" )->Text(), "local" ) )
	{
	    return;
	}

	strbufp = new StrBuf;
	strbufp->Set( varList->GetVar( "name" ) );

	depots->Put( strbufp );
}

void ClientUserDirs::SetDirs( StrBufArray *pStrBufArray )
{
	dirs = pStrBufArray;
}

void ClientUserDirs::OutputStat( StrDict *varList )
{
	StrBuf *strbufp;

	strbufp = new StrBuf;
	strbufp->Set( varList->GetVar( "dir" ) );

	dirs->Put( strbufp );
}

void ClientUserFstat::SetDepotfiles( StrBufArray *pStrBufArray )
{
	depotfiles = pStrBufArray;
}

void ClientUserFstat::SetDirs( StrBufArray *pStrBufArray )
{
	dirs = pStrBufArray;
}

void ClientUserFstat::HandleError( Error *err )
{
	if( err->GetId( 0 )->UniqueCode() == 0x1972 )
	{
	    /*
	     * Error encountered was "protected namespace - access denied",
	     * which is expected when no access is granted to a directory
	     * but access is granted to a subdirectory at some lower
	     * level in the hierarchy. Ignore the "error" by
	     * returning without catching the error.
	     */
	    return;
	}
	else if( err->GetId( 0 )->UniqueCode() == 0x1977 )
	{
	    /*
	     * Error encountered was "no such file(s)", which is expected when
	     * there are only subdirectories (that is, no depot files) in a
	     * directory. Ignore the "error" by returning without catching
	     * the error.
	     */
	    return;
	}

	/*
	 * Let a superclass catch the error.
	 */
	ClientUserCatchError::HandleError( err );
}

void ClientUserFstat::OutputStat( StrDict *varList )
{
	StrPtr *strptrp;
	StrBuf *strbufp;

	strbufp = new StrBuf;

	if( strptrp = varList->GetVar( "depotFile" ) )
	{
	    strbufp->Set( strptrp );
	    depotfiles->Put( strbufp );
	}

	if( strptrp = varList->GetVar( "dir" ) )
	{
	    strbufp->Set( strptrp );
	    dirs->Put( strbufp );
	}
}

void ClientUserFilelog::HandleError( Error *err )
{
	if( err->GetId( 0 )->UniqueCode() == 0x1972 )
	{
	    /*
	     * Error encountered was "protected namespace - access denied",
	     * which is expected when no access is granted to a directory
	     * but access is granted to a subdirectory at some lower
	     * level in the hierarchy. Ignore the "error" by
	     * returning without catching the error.
	     */
	    return;
	}
	else if( err->GetId( 0 )->UniqueCode() == 0x1977 )
	{
	    /*
	     * Error encountered was "no such file(s)", which is expected when
	     * there are only subdirectories (that is, no depot files) in a
	     * directory. Ignore the "error" by returning without catching
	     * the error.
	     */
	    return;
	}

	/*
	 * Let a superclass catch the error.
	 */
	ClientUserCatchError::HandleError( err );
}

void setup( OptionsBrowseChild *opts, ClientApiBrowseChild *client, Error *e )
{
	ClientUserClient ui;

	char o[] = "-o";
	char i[] = "-i";

	/*
	 * The argument for the client command is either "-o" or "-i".
	 */
#define NCLIENTARGS 1
	char *op = &o[ 0 ];	// -o
	char *ip = &i[ 0 ];	// -i

	/*
	 * Set up the client API, which includes making the initial connection
	 * to the server.
	 */
	client->Setup( opts->GetPort(), e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Set the name of the user and the client.
	 */
	client->SetUser( opts->GetUsername() );
	client->SetClient( opts->GetClientname() );

	/*
	 * Run the 'client -o' command.
	 */
	ui.Run( opts, client, "client", NCLIENTARGS, &op, e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Run the 'client -i' command.
	 */
	ui.Run( opts, client, "client", NCLIENTARGS, &ip, e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Set the seed of the pseudo-random number generator.
	 */
	srandomL( opts->GetSeed() );

	/*
	 * Determine if directories are to be retrieved using fstat -Dc
	 * instead of a separate dirs command.
	 */
	opts->SetFstatDc( client->GetProtocol( "server" ) );
}

bool browsedone( int ndirs, int ndepotfiles )
{
	if( !ndirs )
	{
	    /*
	     * No subdirectories are present at this level, so it's not
	     * possible to browse any farther down on this depot path.
	     */
	    return true;
	}

	if( ndepotfiles )
	{
	    /*
	     * At least one depot file is present at this level.
	     *
	     * A depot file at this level will suffice 10% of the time, on
	     * average. Generate a random number to determine if a depot
	     * file at this level will suffice.
	     */
	    if( !nrandom( 0, 10 ) )
	    {
		/*
		 * A depot file at this level will suffice, so there's no
		 * need to browse any farther down the repository structure.
		 */
		return true;
	    }
	}

	/*
	 * Continue browsing down the repository structure.
	 */
	return false;
}

void browse( OptionsBrowseChild *opts,
	ClientApiBrowseChild *client, StrBuf *dir, Error *e )
{
	ClientUserDirs uidirs;
	ClientUserFstat uifstat;
	ClientUserFilelog uifilelog;

	StrBufArray dirs;
	StrBufArray depotfiles;

	int ndepotfiles;

	char C[] = "-C";
	char P[] = "-P";
	char Olh[] = "-Olh";
	char Dc[] = "-Dc";
	char L[] = "-L";

	/*
	 * The arguments for the dirs command are "-C" and
	 * the depot path, which is assigned later.
	 */
#define NDIRSARGS 2
	char *dirsargs[ NDIRSARGS ];
	dirsargs[ 0 ] = &C[ 0 ];	// -C

	/*
	 * The arguments for the fstat command are at a minimum
	 * "-P", "-C", and "-Olh". The depot path is assigned later.
	 */
#define MAXFSTATARGS 5
	char *fstatargs[ MAXFSTATARGS ];
	fstatargs[ 0 ] = &P[ 0 ];	// -P
	fstatargs[ 1 ] = &C[ 0 ];	// -C
	fstatargs[ 2 ] = &Olh[ 0 ];	// -Olh
	int ifstatargs = 3;
	int nfstatargs;

	/*
	 * The arguments for the filelog command are "-L" and
	 * the depot file, which is assigned later.
	 */
#define NFILELOGARGS 2
	char *filelogargs[ NFILELOGARGS ];
	filelogargs[ 0 ] = &L[ 0 ];	// -L

	if( opts->FstatDc() )
	{
	    /*
	     * Get directories using fstat -Dc instead of a separate dirs
	     * command; append -Dc to the fstat arguments and set where
	     * the directories returned are collected.
	     */
	    fstatargs[ ifstatargs++ ] = &Dc[ 0 ];	// -Dc
	    uifstat.SetDirs( &dirs );
	}
	else
	{
	    /*
	     * Get directories using a separate dirs command;
	     * set where the directories returned are collected.
	     */
	    uidirs.SetDirs( &dirs );
	}

	nfstatargs = ifstatargs + 1;

	/*
	 * Set where the depot files returned are collected.
	 */
	uifstat.SetDepotfiles( &depotfiles );

	while( true )
	{
	    dir->Append( "/*" );

	    dirs.Clear();
	    depotfiles.Clear();

	    if( !opts->FstatDc() )
	    {
		/*
		 * Get directories using a separate dirs command;
		 * run the 'dirs -C <depot-path>' command.
		 */
		dirsargs[ 1 ] = dir->Text();
		uidirs.Run( opts, client, "dirs",
		    NDIRSARGS, &dirsargs[ 0 ], e );
		if( e->Test() )
		{
		    return;
		}
	    }

	    /*
	     * Run the 'fstat -P -C -Olh [-Dc] <depot-path>' command.
	     */
	    fstatargs[ ifstatargs ] = dir->Text();
	    uifstat.Run( opts, client, "fstat",
		nfstatargs, &fstatargs[ 0 ], e );
	    if( e->Test() )
	    {
		return;
	    }

	    ndepotfiles = depotfiles.Count();
	    if( browsedone( dirs.Count(), ndepotfiles ) )
	    {
		/*
		 * We're done browsing down the repository structure
		 * for this browse.
		 */
		break;
	    }

	    /*
	     * Get a random subdirectory to browse.
	     */
	    dir->Set( dirs.GetRandom() );
	}

	/*
	 * Run the 'filelog -L <depot-file>' command.
	 */
	if( ndepotfiles )
	{
	    /*
	     * At least one depot file is present at this level;
	     * get a random filename at this level for which the
	     * 'filelog -L <depot-file>' command will be run.
	     */
	    filelogargs[ 1 ] = depotfiles.GetRandom()->Text();
	}
	else
	{
	    /*
	     * No depot files exist at this level; use the path as the
	     * argument for the 'filelog -L <depot-file>' command (which
	     * will return nothing) so that at least the browse gets credited.
	     */
	    filelogargs[ 1 ] = dir->Text();
	}
	uifilelog.Run( opts, client, "filelog",
	    NFILELOGARGS, &filelogargs[ 0 ], e );
	if( e->Test() )
	{
	    return;
	}
}

void runme( OptionsBrowseChild *opts, ClientApiBrowseChild *client, Error *e )
{
	ClientUserDepots uidepots;

	StrBufArray depots;

	int ibrowse;

	StrBuf dir;

	/*
	 * Set where the depots returned are collected.
	 */
	uidepots.SetDepots( &depots );

	/*
	 * Run the 'depots' command.
	 */
	uidepots.Run( opts, client, "depots", e );
	if( e->Test() )
	{
	    return;
	}

	for( ibrowse = opts->GetBrowses(); ibrowse > 0; ibrowse-- )
	{
	    /*
	     * Get a random depot to browse.
	     */
	    dir.Set( "//" );
	    dir.Append( depots.GetRandom() );

	    client->NewBrowse();	// set newbrowse to true

	    browse( opts, client, &dir, e );
	    if( e->Test() )
	    {
		return;
	    }
	}
}

void cleanup( OptionsBrowseChild *opts, ClientApiBrowseChild *client, Error *e )
{
	ClientUserQuiet ui;

	char d[] = "-d";

	/*
	 * The arguments for the client and user commands are "-d" and
	 * the name of the client or user, which is assigned later.
	 */
#define NDARGS 2
	char *dargs[ NDARGS ];
	dargs[ 0 ] = &d[ 0 ];	// -d

	/*
	 * Run the 'client -d <client-name>' command.
	 */
	dargs[ 1 ] = client->GetClient().Text();
	ui.Run( opts, client, "client", NDARGS, &dargs[ 0 ], e );
	if( e->Test() )
	{
	    return;
	}

	/*
	 * Run the 'user -d <user-name>' command.
	 */
	dargs[ 1 ] = client->GetUser().Text();
	ui.Run( opts, client, "user", NDARGS, &dargs[ 0 ], e );
	if( e->Test() )
	{
	    return;
	}

	client->Cleanup( e );
	if( e->Test() )
	{
	    return;
	}
}

int main( int argc, char **argv );
int main( int argc, char **argv )
{
	ErrorLog el;
	Error e;

	OptionsBrowseChild opts;

	ClientApiBrowseChild client;

	/*
	 * Change the standard tag for reported messages.
	 */
	el.SetTag( "BrowseChild" );

	/*
	 * Skip over the program name and parse any arguments. If unexpected
	 * arguments were specified, abort with a usage message.
	 */
	argc--;
	argv++;
	opts.Parse( argc, argv, "c:dn:p:s:u:V", OPT_NONE,
	    MsgBrowseChild::ShortUsage, &e );
	el.Abort( &e );

	if( opts[ 'V' ] )
	{
	    /*
	     * Print version information and exit.
	     */
	    StrBuf s;
	    ident.GetMessage( &s );
	    printf( "%s", s.Text() );
	    return 0;
	}

	/*
	 * Set options using the values of any specified arguments.
	 */
	opts.SetOpts( &e );
	if( e.Test() )
	{
	    e.Set( MsgBrowseChild::ShortUsage );
	    el.Abort( &e );
	}

	/*
	 * Set up for browsing the repository structure.
	 */
	setup( &opts, &client, &e );
	if( e.Test() )
	{
	    el.Abort( &e );
	}

	/*
	 * Browse the repository structure some number of times.
	 */
	runme( &opts, &client, &e );
	if( e.Test() )
	{
	    el.Abort( &e );
	}

	/*
	 * Clean up after browsing the repository structure.
	 */
	cleanup( &opts, &client, &e );
	if( e.Test() )
	{
	    el.Abort( &e );
	}

	return 0;
}
